/*
   Copyright(c) 2002-2011 Thetus Corporation. www.thetus.com

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
*/

package com.thetus.oss.build.core.util;

import org.apache.commons.io.FilenameUtils;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;

import java.io.File;
import java.util.ArrayList;
import java.util.List;
import java.util.logging.Logger;

/**
 * <p>
 * This class is intended to provide a simple interface for managing local configuration overrides.
 * It will probably be used most frequently in maven plugins or other build/compile-time tools.
 * This class allows us to search an ordered list of directories when looking
 * for a config file, and can be useful for any situation in which you may have global settings with
 * local overrides.
 * </p>
 * <p>
 * The main purpose of the class is to maintain an ordered list of directories which will be searched
 * for local config files.  By default, this list includes two directories:
 * </p>
 * <ol>
 *     <li>&lt;working-dir&gt;/src/main/local-config</li>
 *     <li>~/.m2/&lt;settings-dir&gt;/&lt;config-group-name&gt;/local-config</li>
 * </ol>
 * <p>
 * The "preferred" local config override directory is &lt;working-dir&gt;/src/main/local-config.
 * However, if a developer wishes to check out a clean source code tree from the source repository
 * with some regularity, storing local config files inside of the source tree can be problematic.
 * For this reason, we also search a second directory under ~/.m2.  Developers can store config
 * files there without worrying about them being lost when doing a clean source code checkout.
 * </p>
 * <p>
 * Both the <code>working-dir</code> and <code>config-group-name</code> values can be provided
 * as arguments to the constructor when instantiating the class.  The class also provides methods
 * for manually adding additional directories to the list.
 * </p>
 * <p>
 * The two most useful methods on this class are {@link #getLocalConfigDirectories()} and
 * {@link #findLocalConfigFile(String)}. *
 * </p>
 */
public class LocalConfigurationManager {

    // constant to hold the directory name for all of our default local config directories
    public static final String DIR_NAME_LOCAL_CONFIG = "local-config";
    // constant to hold the directory name for the local config directory inside of ~/.m2
    public static String M2_SETTINGS_DIR_NAME = "maven-toolbox";

    // constant to hold the relative path from the base / working dir to the most immediate
    //  local config directory
    public static final String REL_PATH_LOCAL_CONFIG_DIR = "src/main/" + DIR_NAME_LOCAL_CONFIG;

    private final List<File> m_localConfigDirs;
    private final Logger m_logger;


    /**
     * Sets what directory under ~/.m2 to look in for configuration files.
     * @param dirName the directory name to use
     */
    public static void setM2SettingsDirectory ( final @NotNull String dirName ) {
        M2_SETTINGS_DIR_NAME = dirName;
    }
    /**
     * Basic constructor that only requires a configuration "group name" to be passed in.  Assumes
     * that the current working directory should be treated as the base dir.  This constructor passes
     * through to {@link #LocalConfigurationManager(java.io.File, String)}, so please see the docs
     * there for more details.
     * @param configGroupName a name for the group of things that you are configuring.  This will be
     *  used as part of the directory path for some of the local config directories.
     */
    public LocalConfigurationManager(final String configGroupName) {
        this(new File("."), configGroupName);
    }

    /**
     * Constructor that allows user to specify a base / working dir in addition to the configuration
     * group name.  This constructor passes
     * through to {@link #LocalConfigurationManager(java.io.File, String, java.util.logging.Logger)},
     * so please see the docs there for more details.
     * @param workingDir the base / working dir to use when building the default list of directories
     * @param configGroupName a name for the group of things that you are configuring.  This will be
     *  used as part of the directory path for some of the local config directories.
     */
    public LocalConfigurationManager(final File workingDir, final String configGroupName) {
        this(workingDir, configGroupName, Logger.getLogger(""));
    }

    /**
     * The main constructor, which is called by all of the other constructors.
     * @param workingDir the base / working dir to use when determining the path to the first local
     *  configuration directory.  The first directory in the list will be
     *  &lt;workingDir&gt;/src/main/local-config .
     * @param configGroupName the config "group name", which is used to construct the second local
     *  configuration directory, which will be ~/.m2/&lt;m2settingsDir&gt;/&lt;configGroupName&gt;/local-config .
     * @param logger a logger to write status messages to.  If you've called one of the other
     *  constructors--which don't require this parameter--the default global logger will be used.
     */
    public LocalConfigurationManager(@NotNull final File workingDir,
                                     @NotNull final String configGroupName,
                                     @NotNull final Logger logger) {
        this(workingDir,configGroupName,"",logger);
    }

    /**
     * The main constructor, which is called by all of the other constructors.
     * @param workingDir the base / working dir to use when determining the path to the first local
     *  configuration directory.  The first directory in the list will be
     *  &lt;workingDir&gt;/src/main/local-config .
     * @param configGroupName the config "group name", which is used to construct the second local
     *  configuration directory, which will be ~/.m2/&lt;m2settingsDir&gt;/&lt;configGroupName&gt;/local-config .
     * @param logger a logger to write status messages to.  If you've called one of the other
     *  constructors--which don't require this parameter--the default global logger will be used.
     * @param m2settingsDir the directory under ~/.m2/ to search for config files
     */
    public LocalConfigurationManager(@NotNull final File workingDir,
                                     @NotNull final String configGroupName,
                                     @NotNull final String m2settingsDir,
                                     @NotNull final Logger logger) {
        m_localConfigDirs = new ArrayList<File>() {{
            // add "workingDir/src/main/local-config".
                add(new File(FilenameUtils.concat(workingDir.getAbsolutePath(), REL_PATH_LOCAL_CONFIG_DIR)));
            // add "~/.m2/thetus/configGroupName/local-config".
                add(new File(
                         FilenameUtils.concat(M2LocationUtils.getM2Directory().getAbsolutePath(),
                                 (m2settingsDir.isEmpty() ? M2_SETTINGS_DIR_NAME : m2settingsDir) +
                                         File.separator + configGroupName + File.separator +
                                         DIR_NAME_LOCAL_CONFIG)));
        }};
        m_logger = logger;
        // print out a simple log message indicating what our directories are
        logLocalConfigDirs();
    }

    /**
     * Get a list (sorted by precedence) of the local configuration directories that should be searched
     * for configuration files.
     * @return a sorted list of directories to search for local configuration files
     */
    public List<File> getLocalConfigDirectories() {
        return m_localConfigDirs;
    }

    /**
     * Add an additional local configuration directory to the list of directories that should be
     * searched.  The specified directory will be pre-pended to the existing search list, meaning
     * that it will have the highest priority when search for local configuration files.
     * @param additionalConfigDir the new directory to add to the list of config dirs
     */
    public void addDirectory(File additionalConfigDir) {
        m_logger.info("Adding preferred local config directory: '" + additionalConfigDir.getAbsolutePath() + "'");
        m_localConfigDirs.add(0, additionalConfigDir);
    }

    /**
     * Searches for the specified config file in the list of local config directories.  The
     * directories are searched in order, and once a file is found that matches the specified name,
     * the method returns that file without searching any of the remaining directories.
     * @param configFileName the name of the local config file to search the config directories for
     * @return the first matching config file found in the ordered list of directories, or
     *  <code>null</code> if no file was found.
     */
    public @Nullable File findLocalConfigFile(final String configFileName) {
        for (final File localConfigDir : m_localConfigDirs) {
            final File localConfigFile = new File(
                    FilenameUtils.concat(localConfigDir.getAbsolutePath(),
                            configFileName));
            if (localConfigFile.exists()) {
                m_logger.info("Found local config file: '" + localConfigFile.getAbsolutePath() + "'");
                return localConfigFile;
            } else {
                m_logger.info("No local config file found at: '" + localConfigFile.getAbsolutePath() + "'");
            }
        }
        return null;
    }

    /**
     * Searches for the specified config file in the list of local config directories.  The
     * directories are searched in order, and config files are added to the list as they are
     * found. The files will be returned with the highest priority files earliest in the list.
     * @param configFileName the name of the local config file to search the config directories for
     * @return a list of all config files with the appropriate name, ordered by priority, with
     * the highest priority files appearing first in the list.
     */
    public List<File> findAllLocalConfigFile(final String configFileName) {
        List<File> ret = new ArrayList<File>();
        for (final File localConfigDir : m_localConfigDirs) {
            final File localConfigFile = new File(
                    FilenameUtils.concat(localConfigDir.getAbsolutePath(),
                            configFileName));
            if (localConfigFile.exists()) {
                m_logger.info("Found local config file: '" + localConfigFile.getAbsolutePath() + "'");
                ret.add(localConfigFile);
            }
        }
        return ret;
    }

    /**
     * Simple utility method that logs the list of local config directories in their order of
     * precedence.
     */
    public void logLocalConfigDirs() {
        StringBuilder msg = new StringBuilder().append("Local config dirs:\n");

        for (File configDir : m_localConfigDirs) {
            msg.append("\t").append(configDir.getAbsolutePath()).append("\n");
        }
        m_logger.info(msg.toString());
    }

}